// Linked lists of strings

#pragma once

#include <memory>
#include <string>

namespace string_list {

// A list is made out of Cons_cells, each of which contains one string and a
// shared_ptr to the rest of the list.
struct Cons_cell
{
    std::string first;
    std::shared_ptr<Cons_cell> rest;
};

// A list is then a shared_ptr to a Cons_cell:
using List = std::shared_ptr<Cons_cell>;

// Constructs a new cons cell on the free store, returning a shared pointer
// to it.
inline List cons(const std::string first, const List& rest)
{
    return std::make_shared<Cons_cell>(Cons_cell{first, rest});
}

// The empty list.
extern const List string_empty;

// Returns whether the given list is empty (nullptr).
bool is_empty(List);

// Returns whether the given list is non-empty.
bool is_cons(List);

// Returns the length of the given list.
size_t length(List);

// Returns a list containing the elements of `before` followed by the
// elements of `after`, the result sharing storage with `after`.
List append(List before, List after);

// Concatenates `before` onto `after` by destructively updating the final
// `rest` of `before`; if `before` is `nullptr`, it becomes a copy of `after`.
void concat(List& before, List after);

// Returns the reverse of a list.
List reverse(List);

} // end namespace string_list
